/*******************************************************************************

  Intel Data Center Bridging (DCB) Software
  Copyright(c) 2007-2008 Intel Corporation.

  This program is free software; you can redistribute it and/or modify it
  under the terms and conditions of the GNU General Public License,
  version 2, as published by the Free Software Foundation.

  This program is distributed in the hope it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.

  The full GNU General Public License is included in this distribution in
  the file called "COPYING".

  Contact Information:
  e1000-eedc Mailing List <e1000-eedc@lists.sourceforge.net>
  Intel Corporation, 5200 N.E. Elam Young Parkway, Hillsboro, OR 97124-6497

*******************************************************************************/

#include <errno.h>
#include <unistd.h>
#include "gn_test.h"

#define MAX_ADDR_LEN 32  /* instead of including linux/netdevice.h */

static int init_socket(void)
{
	int sd;
	int rcv_size;
	struct sockaddr_nl snl;

	sd = socket(AF_NETLINK, SOCK_RAW, NETLINK_GENERIC);
	if (sd < 0)
		return sd;
  
	if (setsockopt(sd, SOL_SOCKET, SO_RCVBUF, &rcv_size, sizeof(int)) < 0) {
		close(sd);
		return -EIO;
	}
  
	memset((void *)&snl, 0, sizeof(struct sockaddr_nl));
	snl.nl_family = AF_NETLINK;
	snl.nl_pid = getpid();
  
	if (bind(sd, (struct sockaddr *)&snl, sizeof(struct sockaddr_nl)) < 0) {
		close(sd);
		return -EIO;
	}
  
	return sd;
}

static struct nlmsghdr *start_msg(__u16 family_id, __u8 genl_cmd)
{
	struct nlmsghdr *nlh;
	struct genlmsghdr *glh;

	nlh = (struct nlmsghdr *)malloc(MAX_MSG_SIZE);
	if (NULL==nlh)
		return NULL;
	memset((void *)nlh, 0, MAX_MSG_SIZE);
	nlh->nlmsg_len = NLMSG_LENGTH(GENL_HDRLEN);
	nlh->nlmsg_type = family_id;
	nlh->nlmsg_flags = NLM_F_REQUEST;
	nlh->nlmsg_seq = 0;
	nlh->nlmsg_pid = getpid();
	glh = NLMSG_DATA(nlh);
	glh->cmd = genl_cmd;
	glh->version = 1;

	return nlh;
}

static struct nlattr *add_nla(struct nlmsghdr *nlh, __u16 nla_type,
                              void *attr, __u16 nla_len)
{
	struct nlattr *nla;

	nla = (struct nlattr *)((char *)nlh + nlh->nlmsg_len);
	nla->nla_type = nla_type;
	nla->nla_len = nla_len + NLA_HDRLEN;
	if (attr)
		memcpy(NLA_DATA(nla), attr, nla_len);
	nlh->nlmsg_len += NLMSG_ALIGN(nla->nla_len);

	return nla;
}

static int send_msg(struct nlmsghdr *nlh)
{
	struct sockaddr_nl nladdr;
	void *buf = (void *)nlh;
	int r, len = nlh->nlmsg_len;

	memset(&nladdr, 0, sizeof(nladdr));
	nladdr.nl_family = AF_NETLINK;
	
	do {
		r = sendto(nl_sd, buf, len, 0, (struct sockaddr *)&nladdr,
			sizeof(nladdr));
	} while (r < 0 && errno == EINTR);


	if (r < 0)
		return 1;
	else
		return 0;
}

static struct nlmsghdr *get_msg(void)
{
	struct nlmsghdr *nlh;
	int len;

	nlh = (struct nlmsghdr *)malloc(MAX_MSG_SIZE);
	if (NULL==nlh)
		return NULL;
	memset(nlh, 0, MAX_MSG_SIZE);

	len = recv(nl_sd, (void *)nlh, MAX_MSG_SIZE, 0);

	if ((nlh->nlmsg_type == NLMSG_ERROR) || (len < 0) ||
	    !(NLMSG_OK(nlh, len))) {
		free(nlh);
		return NULL;
	}

	return nlh;
}
  
static int get_gnl_id(void)
{
	struct nlmsghdr *nlh;
	struct nlattr *nla;

	nlh = start_msg(GENL_ID_CTRL, CTRL_CMD_GETFAMILY);
	if (NULL==nlh)
		return -EIO;

	add_nla(nlh, CTRL_ATTR_FAMILY_NAME, family,
	        strlen(IXGBE_NLTYPE_DCB_NAME) + 1);

	if (send_msg(nlh)) {
		printf("family error 1\n");
		return -EIO;
	}

	nlh = get_msg();
	if (!nlh) {
		printf("family error 2\n");
		return -EIO;
	}

	nla = (struct nlattr *)GENLMSG_DATA(nlh);

	/* genetlink puts family name first, so go to next attribute.
	 * This is where a full genetlink library would be useful.
	 */
	nla = (struct nlattr *)((char *)nla + NLA_ALIGN(nla->nla_len));

	if (nla->nla_type != CTRL_ATTR_FAMILY_ID) {
		printf("family error 3\n");
		return -EIO;
	}

	fid = *(__u16 *)NLA_DATA(nla);

	return 0;
}

static int get_state(char *ifname, __u8 *state)
{
	struct nlmsghdr *nlh;
	struct genlmsghdr *glh;
	struct nlattr *nla;

	nlh = start_msg(fid, DCB_C_GSTATE);
	if (NULL==nlh)
		return -EIO;

	add_nla(nlh, DCB_A_IFNAME, (void *)ifname,
	        strlen(IXGBE_NLTYPE_DCB_NAME) + 1);

	if (send_msg(nlh))
		return -EIO;

	nlh = get_msg();
	if (!nlh)
		return -EIO;

	glh = (struct genlmsghdr *)NLMSG_DATA(nlh);
	nla = (struct nlattr *)GENLMSG_DATA(nlh);

	if (glh->cmd != DCB_C_GSTATE) {
		printf("Hmm, this is not the message we were expecting.\n");
		return -EIO;
	}
	if (nla->nla_type != DCB_A_STATE) {
		/* Do we really want to code up an attribute parser?? */
		printf("A full libnetlink (with genl and attribute support) "
		       "would sure be nice.\n");
		return -EIO;
	}
	*state = *(__u8 *)NLA_DATA(nla);

	return 0;
}
 

static int get_link_spd(char *ifname, int *speed)
{
	struct nlmsghdr *nlh;
	struct genlmsghdr *glh;
	struct nlattr *nla;

	nlh = start_msg(fid, DCB_C_GLINK_SPD);
	if (NULL==nlh)
		return -EIO;

	add_nla(nlh, DCB_A_IFNAME, (void *)ifname,
	        strlen(IXGBE_NLTYPE_DCB_NAME) + 1);

	if (send_msg(nlh))
		return -EIO;

	nlh = get_msg();
	if (!nlh)
		return -EIO;

	glh = (struct genlmsghdr *)NLMSG_DATA(nlh);
	nla = (struct nlattr *)GENLMSG_DATA(nlh);

	if (glh->cmd != DCB_C_GLINK_SPD) {
		printf("Hmm, this is not the message we were expecting.\n");
		return -EIO;
	}
	if (nla->nla_type != DCB_A_LINK_SPD) {
		/* Do we really want to code up an attribute parser?? */
		printf("A full libnetlink (with genl and attribute support) "
		       "would sure be nice.\n");
		return -EIO;
	}
	*speed = *(__u8 *)NLA_DATA(nla);

	return 0;
}
 

static int set_link_spd(char *ifname, __u8 speed)
{
	struct nlmsghdr *nlh;

	nlh = start_msg(fid, DCB_C_SLINK_SPD);
	if (NULL==nlh)
		return -EIO;

	add_nla(nlh, DCB_A_IFNAME, (void *)ifname,
	        strlen(IXGBE_NLTYPE_DCB_NAME) + 1);
	add_nla(nlh, DCB_A_LINK_SPD, (void *)&speed, sizeof(__u8));

	return send_msg(nlh);
}

static int get_pfc_cfg(char *ifname, __u8 *pfc)
{
	struct nlmsghdr *nlh;
	struct genlmsghdr *glh;
	struct nlattr *nla_parent, *nla_child;
	int i;

	nlh = start_msg(fid, DCB_C_PFC_GCFG);
	if (NULL==nlh)
		return -EIO;

	add_nla(nlh, DCB_A_IFNAME, (void *)ifname,
	        strlen(IXGBE_NLTYPE_DCB_NAME) + 1);
	nla_parent = add_nla(nlh, DCB_A_PFC_CFG, NULL, 0);
	nla_child = add_nla(nlh, PFC_A_UP_ALL, NULL, 0);
	nla_parent->nla_len += NLA_ALIGN(nla_child->nla_len);
	if (send_msg(nlh))
		return -EIO;

	nlh = get_msg();
	if (!nlh)
		return -EIO;

	glh = (struct genlmsghdr *)NLMSG_DATA(nlh);
	nla_parent = (struct nlattr *)GENLMSG_DATA(nlh);

	if (glh->cmd != DCB_C_PFC_GCFG) {
		printf("Hmm, this is not the message we were expecting.\n");
		return -EIO;
	}
	if (nla_parent->nla_type != DCB_A_PFC_CFG) {
		/* Do we really want to code up an attribute parser?? */
		printf("A full libnetlink (with genl and attribute support) "
		       "would sure be nice.\n");
		return -EIO;
	}
	nla_child = NLA_DATA(nla_parent);
	nla_parent = (struct nlattr *)((char *)nla_parent +
	                               NLA_ALIGN(nla_parent->nla_len));
	for (i = 0; nla_parent > nla_child; i++) {
		if (i == 8) {
			printf("pfc array out of range\n");
			break;
		}
		pfc[i] = *(__u8 *)NLA_DATA(nla_child);
		nla_child = (struct nlattr *)((char *)nla_child +
		                              NLA_ALIGN(nla_child->nla_len));
	}
	if (nla_parent != nla_child)
		printf("nla pointers are off\n");

	return 0;
}


static int get_pg(char *ifname, struct tc_config *tc, __u8 *bwg, int cmd)
{
	struct nlmsghdr *nlh;
	struct genlmsghdr *glh;
	struct nlattr *class_parent, *param_parent, *nla_child;
	__u8 *p = (__u8 *)tc;
	int i, j;

	nlh = start_msg(fid, cmd);
	if (NULL==nlh)
		return -EIO;

	add_nla(nlh, DCB_A_IFNAME, (void *)ifname,
	        strlen(IXGBE_NLTYPE_DCB_NAME) + 1);
	class_parent = add_nla(nlh, DCB_A_PG_CFG, NULL, 0);
	param_parent = add_nla(nlh, PG_A_TC_ALL, NULL, 0);
	nla_child = add_nla(nlh, TC_A_PARAM_ALL, NULL, 0);
	param_parent->nla_len += NLA_ALIGN(nla_child->nla_len);
	class_parent->nla_len += NLA_ALIGN(param_parent->nla_len);

	nla_child = add_nla(nlh, PG_A_BWG_ALL, NULL, 0);
	class_parent->nla_len += NLA_ALIGN(nla_child->nla_len);
	if (send_msg(nlh))
		return -EIO;

	nlh = get_msg();
	if (!nlh)
		return -EIO;

	glh = (struct genlmsghdr *)NLMSG_DATA(nlh);
	class_parent = (struct nlattr *)GENLMSG_DATA(nlh);

	if (glh->cmd != cmd) {
		printf("Hmm, this is not the message we were expecting.\n");
		return -EIO;
	}
	if (class_parent->nla_type != DCB_A_PG_CFG) {
		/* Do we really want to code up an attribute parser?? */
		printf("A full libnetlink (with genl and attribute support) "
		       "would sure be nice.\n");
		return -EIO;
	}
	param_parent = NLA_DATA(class_parent);
	class_parent = (struct nlattr *)((char *)class_parent +
	                                 NLA_ALIGN(class_parent->nla_len));
	for (i = 0; class_parent > param_parent; i++) {
		if (param_parent->nla_type >= PG_A_TC_0 &&
			param_parent->nla_type < PG_A_TC_MAX) {
			nla_child = NLA_DATA(param_parent);
			param_parent = (struct nlattr *)((char *)param_parent +
			      NLA_ALIGN(param_parent->nla_len));

			for (j = 0; param_parent > nla_child; j++) {
				if (j == TC_A_PARAM_MAX -
					TC_A_PARAM_STRICT_PRIO) {
					printf("parameter array out of "
						"range\n");
					break;
				}
				*p = *(__u8 *)NLA_DATA(nla_child);
				nla_child =
					(struct nlattr *)((char *)nla_child +
					NLA_ALIGN(nla_child->nla_len));
				p++;
			}
			if (param_parent != nla_child)
				printf("param_parent and nla_child pointers "
					"are off\n");
		} else if (param_parent->nla_type >= PG_A_BWG_0 &&
			param_parent->nla_type < PG_A_BWG_MAX) {
			j = param_parent->nla_type - PG_A_BWG_0;
			bwg[j] = *(__u8 *)NLA_DATA(param_parent);
			param_parent = (struct nlattr *)((char *)param_parent +
				NLA_ALIGN(param_parent->nla_len));
		}
		else
			printf("unknown param_parent type = %d\n",
				param_parent->nla_type);

	}

	if (class_parent != param_parent)
		printf("class_parent and param_parent pointers are off\n");

	return 0;
}

static int get_perm_hwaddr(char *ifname, __u8 *buf)
{
	struct nlmsghdr *nlh;
	struct genlmsghdr *glh;
	struct nlattr *nla_parent, *nla_child;
	int i;

	nlh = start_msg(fid, DCB_C_GPERM_HWADDR);
	if (NULL==nlh)
		return -EIO;

	add_nla(nlh, DCB_A_IFNAME, (void *)ifname,
	        strlen(IXGBE_NLTYPE_DCB_NAME) + 1);
	nla_parent = add_nla(nlh, DCB_A_PERM_HWADDR, NULL, 0);
	nla_child = add_nla(nlh, PERM_HW_A_ALL, NULL, 0);
	nla_parent->nla_len += NLA_ALIGN(nla_child->nla_len);
	if (send_msg(nlh))
		return -EIO;

	nlh = get_msg();
	if (!nlh)
		return -EIO;

	glh = (struct genlmsghdr *)NLMSG_DATA(nlh);
	nla_parent = (struct nlattr *)GENLMSG_DATA(nlh);

	if (glh->cmd != DCB_C_GPERM_HWADDR) {
		printf("Hmm, this is not the message we were expecting.\n");
		return -EIO;
	}
	if (nla_parent->nla_type != DCB_A_PERM_HWADDR) {
		/* Do we really want to code up an attribute parser?? */
		printf("A full libnetlink (with genl and attribute support) "
		       "would sure be nice.\n");
		return -EIO;
	}
	nla_child = NLA_DATA(nla_parent);
	nla_parent = (struct nlattr *)((char *)nla_parent +
	                               NLA_ALIGN(nla_parent->nla_len));

	for (i = 0; nla_parent > nla_child; i++) {
		*(buf+i) = *(__u8 *)NLA_DATA(nla_child);
		nla_child = (struct nlattr *)((char *)nla_child +
		                              NLA_ALIGN(nla_child->nla_len));
	}
	if (nla_parent != nla_child)
		printf("nla pointers are off\n");

	return 0;
}
 
/*
static int set_hw_all(char *device_name)
{
	struct nlmsghdr *nlh;
	int status = 1;

	nlh = start_msg(fid, DCB_C_SET_ALL);
	if (NULL==nlh)
		return -EIO;

	add_nla(nlh, DCB_A_IFNAME, (void *)device_name,
	        strlen(IXGBE_NLTYPE_DCB_NAME) + 1);
	add_nla(nlh, DCB_A_SET_ALL, (void *)&status, sizeof(__u8));

	return send_msg(nlh);
}
*/

int main(int argc, char *argv[])
{
	struct tc_config tc[8];
	int i, err = 0;
	int speed;
	__u8 state;
	__u8 pfc[8];
	__u8 bwg[8];
	__u8 mac[MAX_ADDR_LEN];

	if (argc != 2 && argc != 3) {
		fprintf(stderr, "usage: %s <ifname> [link speed]\n", argv[0]);
		exit(1);
	}

	if ((nl_sd = init_socket()) < 0) {
		fprintf(stderr, "Error creating Netlink socket\n");
		return nl_sd;
	}

	err = get_gnl_id();
	if (err) {
		fprintf(stderr, "Error getting family id\n");
		goto err_main;
	}
	printf("IXGBE_DCB generic netlink family ID = %d\n", fid);

	err = get_state(argv[1], &state);
	if (err) {
		fprintf(stderr, "Error getting DCB state\n");
		goto err_main;
	}
/*
	if (!state) {
		err = set_state(argv[1], 1);
		if (err)
			goto err_main;
	}
*/
	printf("DCB State is %d\n", state);

	if (argc == 3) {
		speed = atoi(argv[2]);
		if (speed < 0 || speed > 9)
			speed = 0;

		err = set_link_spd(argv[1], speed);
		if (err)
			goto err_main;
	}

	err = get_link_spd(argv[1], &speed);
	if (err) {
		fprintf(stderr, "Error getting DCB link speed\n");
		goto err_main;
	}
	printf("DCB link speed is %d\n", speed);

	get_pfc_cfg(argv[1], pfc);
	printf("\nPFC config:\n");
	for (i=0; i<8; i++) { printf("%d ", pfc[i]); pfc[i] = !pfc[i]; }
	printf("\n");

	get_pg(argv[1], tc, bwg, DCB_C_PGTX_GCFG);
	printf("\nPG config for TX:\n");
	for (i = 0; i < 8; i++) {
		printf("%d: pr=%d\tbwgid=%d\tbw%%=%d\tu2t=%d\tlk%%=%d\n", i,
			tc[i].prio_type,
			tc[i].bwg_id,
			tc[i].bwg_percent,
			tc[i].up_to_tc_bitmap,
			bwg[i]);

		tc[i].prio_type = 3;
		tc[i].bwg_id = i;
		tc[i].bwg_percent = 100;
		tc[i].up_to_tc_bitmap = i;
		bwg[i] = 12 + (i & 1);
	}

/*	printf("set_pfc_cfg = %d\n", set_pfc_cfg(argv[1], pfc)); */
/*	printf("set_rx_pg = %d\n", set_pg(argv[1], tc, bwg, DCB_C_PGRX_SCFG));*/
/*	printf("set_hw_all = %d\n", set_hw_all(argv[1])); */

	memset(bwg, 0, sizeof(bwg));
	memset(&tc[0], 0, sizeof(tc));
	printf("\nPG config for RX:\n");
	get_pg(argv[1], tc, bwg, DCB_C_PGRX_GCFG);
	for (i = 0; i < 8; i++) {
		printf("%d: pr=%d\tbwgid=%d\tbw%%=%d\tu2t=%d\tlk%%=%d\n", i,
			tc[i].prio_type,
			tc[i].bwg_id,
			tc[i].bwg_percent,
			tc[i].up_to_tc_bitmap,
			bwg[i]);
	}

	get_perm_hwaddr(argv[1], &mac[0]);
	printf("Permanent MAC: ");
	for (i = 0; i < 5; i++)
		printf("%02x:", mac[i]);
	printf("%02x\n", mac[i]);

	//get_pfc_stats(nl_sd, id, "eth3");
	//get_pg_stats(nl_sd, id, "eth3");

err_main:
	close(nl_sd);

	return err;
}

