/*******************************************************************************

  Intel Data Center Bridging (DCB) Software
  Copyright(c) 2007-2008 Intel Corporation.

  This program is free software; you can redistribute it and/or modify it
  under the terms and conditions of the GNU General Public License,
  version 2, as published by the Free Software Foundation.

  This program is distributed in the hope it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.

  The full GNU General Public License is included in this distribution in
  the file called "COPYING".

  Contact Information:
  e1000-eedc Mailing List <e1000-eedc@lists.sourceforge.net>
  Intel Corporation, 5200 N.E. Elam Young Parkway, Hillsboro, OR 97124-6497

*******************************************************************************/

#include "dcb_osdep.h"
#include "states.h"
#include "tlv.h"
#include "ports.h"
#include "l2_packet.h"
#include "dcb_types.h"

struct port *porthead = NULL; /* Head pointer */
struct port *portcurrent = NULL; /* Working  pointer loaded from ports or
				  * port->next */


void agent_receive(void *, const u8 *, const u8 *, size_t);

void set_lldp_port_enable_state(const char *ifname, int enable)
{
	struct port *port = NULL;

	port = porthead;
	while (port != NULL) {
		if (!strncmp(ifname, port->ifname,
			MAX_DEVICE_NAME_LEN))  /* Localization OK */
			break;
		port = port->next;
	}

	if (port == NULL) {
		return;
	}

	port->portEnabled = (u8)enable;
	if (enable) {
		port->adminStatus = enabledRxTx;
	} else {
		port->adminStatus = disabled;
		port->portrx.rxInfoAge = FALSE;
	}
	run_tx_sm(port, FALSE);
	run_rx_sm(port, FALSE);

}

int set_port_hw_resetting(const char *ifname, int resetting)
{
	struct port *port = NULL;

	port = porthead;
	while (port != NULL) {
		if (!strncmp(ifname, port->ifname,
			MAX_DEVICE_NAME_LEN)) {  /* Localization OK */
			break;
		}
		port = port->next;
	}

	if (port == NULL) {
		return -1;
	}

	port->hw_resetting = (u8)resetting;

	return port->hw_resetting;
}

int get_port_hw_resetting(const char *ifname)
{
	struct port *port = NULL;

	port = porthead;
	while (port != NULL) {
		if (!strncmp(ifname, port->ifname,
			MAX_DEVICE_NAME_LEN)) {  /* Localization OK */
			break;
		}
		port = port->next;
	}

	if (port)
		return port->hw_resetting;
	else
		return 0;
}


int add_port(const char *ifname)
{
	struct port *newport;
	boolean_t   success;

	newport = porthead;
	while (newport != NULL) {
		if (!strncmp(ifname, newport->ifname,
			MAX_DEVICE_NAME_LEN)) {
			return 0;
		}
		newport = newport->next;
	}

	newport  = (struct port *)malloc(sizeof(struct port));
	if (newport == NULL) {
		printf("new port malloc failed\n");
		goto fail;
	}
	memset(newport,0,sizeof(struct port));
	newport->next = NULL;
	newport->ifname = strdup(ifname);
	if (newport->ifname == NULL) {
		printf("new port name malloc failed\n");
		goto fail;
	}

	newport->l2 = l2_packet_init(newport->ifname, NULL, ETH_P_LLDP,
		rxReceiveFrame, newport, 1);
	if (newport->l2 == NULL) {
		printf("Failed to open register layer 2 access to "
			"ETH_P_LLDP\n");
		goto fail;
	}

	/* Initialize relevent port variables on creation */
	newport->porttx.state  = TX_LLDP_INITIALIZE;
	txInitializeLLDP(newport);
	newport->portrx.state = LLDP_WAIT_PORT_OPERATIONAL;
	rxInitializeLLDP(newport);
	newport->hw_resetting = FALSE;
	newport->portEnabled = FALSE;
	newport->adminStatus = enabledRxTx;
	newport->porttx.txTTL = 0;
	newport->msap.length1 = 0;
	newport->msap.msap1 = NULL;
	newport->msap.length2 = 0;
	newport->msap.msap2 = NULL;
	newport->lldpdu = FALSE;
	newport->dcbdu = FALSE;

	l2_packet_get_port_state(newport->l2, (u8 *)&(newport->portEnabled));

	/* Add TLVs */
	newport->porttlvs.Type1 = construct_type1_tlv(newport);
	if (newport->porttlvs.Type1 == NULL) {
		printf("add_port:  construct_type1_tlv failed\n");
		goto fail_add;
	}

	newport->porttlvs.Type2 = construct_type2_tlv(newport);
	if (newport->porttlvs.Type2 == NULL) {
		printf("add_port:  construct_type2_tlv failed\n");
		goto fail_add;
	}

	newport->porttlvs.Type3 = construct_type3_tlv(newport);
	if (newport->porttlvs.Type3 == NULL) {
		printf("add_port:  construct_type3_tlv failed\n");
		goto fail_add;
	}

	newport->porttlvs.control = construct_control_tlv(newport);
	if (newport->porttlvs.control == NULL) {
		printf("add_port:  construct_control_tlv failed\n");
		goto fail_add;
	}

	newport->porttlvs.pg = construct_pg_tlv(newport, &success);
	if (!success) {
		printf("construct_pg_tlv: failed\n");
		goto fail_add;
	}

	newport->porttlvs.pfc = construct_pfc_tlv(newport, &success);
	if (!success) {
		printf("construct_pfc_tlv: failed\n");
		goto fail_add;
	}

	newport->porttlvs.app = construct_app_tlv(newport,
						FCOE_SUBTYPE, &success);
	if (!success) {
		printf("construct_app_tlv: failed\n");
		goto fail_add;
	}

	newport->porttlvs.Type127 = construct_type127_tlv(newport);
	if (newport->porttlvs.Type127 == NULL) {
		printf("add_port:  construct_type127_tlv failed\n");
		goto fail_add;
	}

	newport->porttlvs.last_peer_tlv = NULL;
	newport->porttlvs.current_peer_tlv = NULL;

	if (porthead) {
		newport->next = porthead;
	}
	porthead = newport;
	return 0;

fail_add:
	newport->porttlvs.last_peer_tlv = NULL;
	newport->porttlvs.current_peer_tlv = NULL;

	if (porthead) {
		newport->next = porthead;
	}
	porthead = newport;
	return -1;

fail:
	if(newport) {
		if(newport->ifname)
			free(newport->ifname);
		free(newport);
	}
	return -1;
}

int remove_port(const char *ifname)
{
	struct port *port = NULL;    /* Pointer to port to remove */
	struct port *parent = NULL;  /* Pointer to previous on port stack */

	port = porthead;
	while (port != NULL) {
		if (!strncmp(ifname, port->ifname,
			MAX_DEVICE_NAME_LEN)) {  /* Localization OK */
			printf("In remove_port: Found port  %s\n",
				port->ifname);
			break;
		}
		parent = port;
		port = port->next;
	}

	if (port == NULL) {
		printf("remove_port: port not present\n");
		return -1;
	}
	/* Close down the socket */
	l2_packet_deinit(port->l2);

	/* Re-initialize relevent port variables */
	port->porttx.state = TX_LLDP_INITIALIZE;
	port->portrx.state = LLDP_WAIT_PORT_OPERATIONAL;
	port->portEnabled  = FALSE;
	port->adminStatus  = disabled;
	port->porttx.txTTL = 0;

	/* Take this port out of the chain */
	if (parent == NULL) {
		porthead = port->next;
	} else if (parent->next == port) { /* Sanity check */
		parent->next = port->next;
	} else {
		return -1;
	}


	/* Remove the tlvs */
	if (port->porttlvs.Type1 != NULL) {
		port->porttlvs.Type1 =
			destruct_unpacked_tlv(port->porttlvs.Type1);
	}

	if (port->porttlvs.Type2 != NULL) {
		port->porttlvs.Type2 =
			destruct_unpacked_tlv(port->porttlvs.Type2);
	}

	if (port->porttlvs.Type3 != NULL) {
		port->porttlvs.Type3 =
			destruct_unpacked_tlv(port->porttlvs.Type3);
	}

	if (port->porttlvs.control != NULL) {
		port->porttlvs.control =
			destruct_unpacked_tlv(port->porttlvs.control);
	}

	if (port->porttlvs.pg != NULL) {
		port->porttlvs.pg = destruct_unpacked_tlv(port->porttlvs.pg);
	}

	if (port->porttlvs.pfc != NULL) {
		port->porttlvs.pfc = destruct_unpacked_tlv(port->porttlvs.pfc);
	}

	if (port->porttlvs.app != NULL) {
		port->porttlvs.app = destruct_unpacked_tlv(port->porttlvs.app);
	}

	if (port->porttlvs.Type127 != NULL) {
		port->porttlvs.Type127 =
			destruct_unpacked_tlv(port->porttlvs.Type127);
	}

	if (port->porttlvs.current_peer_tlv != NULL) {
		port->porttlvs.current_peer_tlv =
			destruct_unpacked_tlv(port->porttlvs.current_peer_tlv);
	}

	if (port->porttlvs.last_peer_tlv != NULL) {
		port->porttlvs.last_peer_tlv =
			destruct_unpacked_tlv(port->porttlvs.last_peer_tlv);
	}

	if (port->msap.msap1) {
		free(port->msap.msap1);
		port->msap.msap1 = NULL;
	}

	if (port->msap.msap2) {
		free(port->msap.msap2);
		port->msap.msap2 = NULL;
	}

	if (port->portrx.framein)
		free(port->portrx.framein);

	if (port->porttx.frameout)
		free(port->porttx.frameout);

	if (port->ifname)
		free(port->ifname);

	free(port);
	return 0;
}
